#! /usr/bin/python
# -*- coding: future_fstrings -*-
"""
Make Blender python modules binaries
"""

# STD_LIB imports
import logging
import os
import pathlib
import platform
import subprocess
import sys
import typing
from typing import Dict, List, Optional, Tuple

# PYPI imports
import cmake
import cmakegenerators

# Relative imports
from bpybuild import BITNESS

LOGGER = logging.getLogger(__name__)

# Platform-specific PYPI imports
try:
    import distro
except ImportError:
    LOGGER.info("Package `distro` not available")

def get_configure_commands(source: pathlib.Path, destination: pathlib.Path,
                           bitness: Optional[int] = None,
                           with_cuda: Optional[bool]=False, 
                           with_optix: Optional[bool] = False,
                           optix_sdk_path: Optional[str] = None) -> List[List[str]]:

    commands = []

    os_configure_args = []

    if bitness is None: bitness = BITNESS

    if platform.system() == "Windows":

        generators = [generator for generator in 
                      cmakegenerators.get_generators() if 
                      generator.name.startswith("Visual Studio")]

        if len(generators) == 0:

            raise Exception("Windows users must have Visual Studio")

        filtered_generators = [generator for generator in generators if 
                                "Visual Studio".casefold() in 
                                generator.name.casefold()]

        if len(filtered_generators) > 0:

            generator_option : str = None

            if BITNESS == 64:

                filtered_generator_options = [option for option in filtered_generators[0].options if "64".casefold() in option.casefold()]

                if len(filtered_generator_options) > 0:

                    generator_option = filtered_generator_options[0]

                else:

                    raise Exception(f"{BITNESS}bit Visual Studio not found, "
                                    f"but Visual Studio is installed. Make "
                                    f"sure you have the correct compilers for "
                                    f"your Python platform ({BITNESS}bit)")

            else:

                generator_option = filtered_generators[0].options[-1]

            if generator_option is None:

                raise Exception(f"No C++ compilers detected on Windows {BITNESS}bit")

            os_configure_args += ["-G", generator_option]

        else:

            raise Exception(f"Visual Studio not found")        

    elif platform.system() == "Linux":

        os_configure_args += ["-DWITH_AUDASPACE=OFF"]

    elif platform.system() == "Darwin":

        os_configure_args += ["-DWITH_OPENMP=OFF", "-DWITH_AUDASPACE=OFF"]

    if platform.system() != "Windows":

        commands.append(["make", "-C", str(source.absolute()), "update"])

    if platform.system() == "Linux":

        pass

    commands.append(['cmake',
                     '-DWITH_PLAYER=OFF', '-DWITH_PYTHON_INSTALL=OFF',
                     '-DWITH_PYTHON_MODULE=ON', 
                     f"-DPYTHON_VERSION={sys.version_info[0]}."
                     f"{sys.version_info[1]}", 
                     "-DWITH_CYCLES_CUDA_BINARIES=ON" if with_cuda else "",
                     "-DWITH_CYCLES_DEVICE_OPTIX=ON" if with_optix else "",
                     f"-DOPTIX_ROOT_DIR={optix_sdk_path}" if 
                     optix_sdk_path is not None else ""] + os_configure_args + 
                    ['-S' + str(source.absolute()), 
                     '-B' + str(destination.absolute())])

    return commands

def get_build_commands(location: pathlib.Path,
                       is_release: Optional[bool] = True) -> List[List[str]]:

    commands = []

    os_build_args = []

    if platform.system() == "Windows": # Windows specific build requirements

        os_build_args += ["--target", "INSTALL", "--config", 
                          f"{'Release' if is_release else 'Debug'}"]

        commands.append(["cmake", "--build", 
                        str(location.absolute())] + os_build_args)

    else:

        commands.append(["make", "-C", str(location.absolute()), "install"])

    return commands

def get_make_commands(source_location: pathlib.Path, 
                      build_location: Optional[pathlib.Path] = None,
                      bitness: Optional[int] = BITNESS,
                      is_release: Optional[bool] = True,
                      with_cuda: Optional[bool]=False, 
                      with_optix: Optional[bool] = False,
                      optix_sdk_path: Optional[str] = None) -> List[List[str]]:

    build_location = build_location if build_location else source_location

    return get_configure_commands(source_location, build_location, bitness, 
                                  with_cuda, with_optix, optix_sdk_path) +\
    get_build_commands(build_location, is_release)



#! /usr/bin/python
# -*- coding: future_fstrings -*-
"""A module for building Blender

This module is mostly meant as an automation tool for building Blender and 
packaging up the module nicely for upload to "bpy-build" later using `twine`
"""

import os
import pathlib
import platform
import struct
import sys

# Monkey-patch 3.4 and below

if sys.version_info < (3,5):

    def home_path() -> pathlib.Path:

        return pathlib.Path(os.path.expanduser("~"))

    pathlib.Path.home = home_path

__all__ = ["make", "sources"]

BITNESS = struct.calcsize("P") * 8



#! /usr/bin/python
# -*- coding: future_fstrings -*-
"""Functions for searching for and retrieving Blender sources and libs
"""

# STD_LIB imports
import abc
import logging
import os
import pkg_resources
import pathlib
import platform
import posixpath
import re
import sys
import typing
from typing import Dict, List, Tuple

# PYPI imports
import git
from git import Repo as GitRepo
from git.exc import NoSuchPathError
from svn.exception import SvnException
from svn.remote import RemoteClient as SvnRepo

# Relative imports
from bpybuild import BITNESS

LOGGER = logging.getLogger(__name__)

# Platform-specific PYPI imports
try:
    import distro
except ImportError as e:
    # Only raise the error if it's Linux environment
    if platform.system == "Linux":
        LOGGER.info("Package `distro` not available")
        raise e

def git_remote_tags(url:str) -> List[Tuple[str, str]]:
    """Get the tags of a remote repository from the repo's `url`
    
    Arguments:
        url {str} -- The repo's address
    """

    return [(tag.split("\t")[0], tag.split("\t")[1]) for tag in 
             git.cmd.Git().ls_remote(url, tags=True).split("\n") if 
             not tag.endswith(r"^{}")]

def git_remote_tagnames(url:str) -> List[str]:
    """Get a list of just names of tags in the remote
    
    Arguments:
        url {str} -- The repo's address
    """

    return [posixpath.basename(tag[1]) for tag in git_remote_tags(url)]

class VersionNotFoundError(Exception):
    """Thrown when `git` or `svn` does not have the specified tag
    
    Arguments:
        Exception {[type]} -- Derived from base exception
    """

    pass

class SourceVersionControl(abc.ABC):
    """Represents code repositories that can be checked out
    """

    @abc.abstractmethod
    def checkout(self, full_path:pathlib.Path):
        """Retrieves the code in an impliementation-specific way

        Must be reimplimented per the code repository type
        """

        raise NotImplementedError

class SvnOSPlatform():

    PYTHON_PATTERN = (r'(?<=python)(\d)(?:\.?)(\d)')
    PYTHON_REGEX = re.compile(PYTHON_PATTERN)

    def __init__(self, svn_url: str):

        self.url : str = svn_url

        self.repo : SvnRepo = SvnRepo(svn_url)

        self.svn_name : str = posixpath.basename(svn_url)

        self.os_name: str = None
        self.os_version : Tuple[int] = None
        self.processor: str = None
        self.bitness: int = None
        self.build_environment: str = None

        if self.svn_name.casefold().startswith("android".casefold()):

            self.os_name = "Android"
        
        elif self.svn_name.casefold().startswith("darwin".casefold()):

            self.os_name = "Darwin"

            darwin_platform_tokens = re.split(r"[\.-]", self.svn_name)

            if len(darwin_platform_tokens) >= 4:

                self.os_version = pkg_resources.parse_version(".".join(darwin_platform_tokens[1: -1]))

                self.processor = darwin_platform_tokens[-1]

        elif self.svn_name.casefold().startswith("linux".casefold()):

            self.os_name = "Linux"

        elif self.svn_name.casefold().startswith("win".casefold()):

            self.os_name = "Windows"

            if "64".casefold() in self.svn_name.casefold():

                self.bitness = 64

            else:

                self.bitness = 32

            if "vc".casefold() in self.svn_name.casefold():

                self.build_environment = "vc" + self.svn_name.split("vc")[-1]

        else:

            self.os_name = None

    def python_versions(self) -> List[Tuple[int, int]]:

        try:

            versions = [re.search(self.PYTHON_REGEX, 
                                  version["name"]) for version in 
                        self.repo.list(extended = True,
                                       rel_path = "python/lib") if
                        re.search(self.PYTHON_REGEX, version["name"]) and 
                        version["kind"] == "file"]

            return [(int(version.group(1)), int(version.group(2))) for version in versions]

        except SvnException:

            return []

class BlenderGit(SourceVersionControl):
    """The Blender `git` sources

    Provides the source code created by Blender Foundation developers.
    """

    BASE_URL = "git://git.blender.org/blender.git"

    # Here I store the Blender `git` sources in a folder in the home directory
    # so that I don't need to waste time constantly pulling from the repository;
    # only minor updates are needed here and there.

    def __init__(self, tag: str = None):

        self.tag = tag

        self.version = pkg_resources.parse_version(tag)

    @classmethod
    def tags(self) -> List[str]:
        """A list of release tags that git is aware of
        
        Returns:
            List[str] -- release tags that are available
        """
        
        return [version for version in git_remote_tagnames(self.BASE_URL) if 
                not str(version).startswith("Studio")] # what is a studio version?

    def checkout(self, full_path: pathlib.Path):
        """Retrieve Blender code from Git
        
        Keyword Arguments:
            full_path {pathlib.Path} -- place to clone code to 
        """

        try:

            repo = GitRepo(str(full_path))

        except:

            GitRepo.clone_from(self.BASE_URL, str(full_path))
            repo = GitRepo(str(full_path))
        
        repo.heads.master.checkout()
        repo.remotes.origin.pull()

        if self.tag is not None:

            repo.git.checkout(self.tag)

        repo.git.submodule('update', '--init', '--recursive')

        for submodule in repo.submodules:

            submodule_repo = submodule.module()
            submodule_repo.heads.master.checkout()
            submodule_repo.remotes.origin.pull()

            if self.tag is not None:

                try:

                    submodule_repo.git.checkout(self.tag)

                except:

                    pass

class BlenderSvn(SourceVersionControl):
    """The Blender `svn` library index

    Provides the set of required libraries to build Blender
    """

    BASE_URL = "https://svn.blender.org/svnroot/bf-blender"
    BASE_REPO = SvnRepo(BASE_URL)

    def __init__(self, svn_url: str):

        self.url = svn_url

        self.repo = SvnRepo(svn_url)

        self.version = pkg_resources.parse_version(os.path.basename(os.path.normpath(svn_url))
                                                   .replace("blender-", "")
                                                   .replace("-release", "")
                                                   .replace("-winfix", "")
                                                   .replace("-", "."))

        self._platforms : List[SvnOSPlatform] = None

        self._platforms_dict : Dict[str, List[Tuple[int, int]]] = None

    @classmethod
    def tags(cls) -> List[str]:
        """The tags that `svn` found
        
        Able to be run as a class method so that we can query `svn` and build
        our list of svn supported versions

        Returns:
            List[str] -- List of strings, the full paths to the svn version
        """

        return [posixpath.join(cls.BASE_URL, "tags", _version) for _version in
                cls.BASE_REPO.list(rel_path="/tags") if 
                _version.startswith("blender")]

    def platforms(self) -> List[SvnOSPlatform]:

        if self._platforms is not None: # Don't perform expensive search twice

            return self._platforms

        results = self.get_platforms(self.url)

        self._platforms = results

        return results

    def platforms_dict(self) -> Dict[str, List[Tuple[int, int]]]:
        """Provides a dictionary of os and Python version compatibility

        Platforms are determined based on the os listed per release of Blender
        API in the `svn` repos
        """

        # SVN search and checkout is expensive; if this seems overkill it's
        # because any other method was taking ~ 5 minutes to list a
        # svn tag!

        if self._platforms_dict is not None: # Don't perform expensive search twice

            return self._platforms_dict

        results = self.get_platforms_dict(self.url)

        self._platforms_dict = results

        return results

    @staticmethod
    def get_platforms(url: str) -> List[SvnOSPlatform]:

        results = []

        try:

            for _os in SvnRepo(url).list(extended = True, rel_path = "lib"):

                if not any([_os["name"] == x for x in ["benchmarks", "package",
                                                       "python", "tests"]]):

                    results.append(SvnOSPlatform(posixpath.join(url + "lib", 
                                                 _os["name"])))

        except SvnException: # This can happen when the "lib" path does not exist

            pass

        return results

    @staticmethod
    def get_platforms_dict(url: str) -> Dict[str, List[Tuple[int, int]]]:

        results = {}

        try:

            _oss = [SvnOSPlatform(posixpath.join(url + "lib", 
                                                 _os["name"])) for
                    _os in SvnRepo(url).list(extended = True, rel_path = "lib") if
                    any([_os["name"].startswith(x) for x in ["android",
                                                             "darwin", "linux",
                                                             "win"]])]

            for _os in _oss:

                results[_os.svn_name] = _os.python_versions()

        except SvnException: # This can happen when the "lib" path does not exist

            pass

        return results

    def checkout(self, full_path: pathlib.Path):
        """Retrieve the `svn` sources
        
        Arguments:
            full_path {pathlib.Path} -- place to put svn sources
        """

        self.repo.checkout(str(full_path))

def git_tags() -> List[BlenderGit]:

    return [BlenderGit(str(tag)) for tag in BlenderGit.tags()]

def svn_tags() -> List[BlenderSvn]:

    return [BlenderSvn(tag_full_path) for tag_full_path in BlenderSvn.tags()]

def get_matched_versions() -> Dict[int, Tuple[List[BlenderGit], 
                                              List[BlenderSvn]]]:
    """Get pairs of sources based on available versions
    
    Returns:
        Dict[int, Tuple[List[BlenderGit],List[BlenderSvn]]]
         -- All versions of Blender and appropriate `BlenderGit` and 
            `BlenderSvn` objects
    """

    result = {}

    gits = git_tags()
    svns = svn_tags()

    for version in set([_svc.version for _svc in gits+svns]):

        result[version] = ([_git for _git in gits if _git.version == version],
                           [_svn for _svn in svns if _svn.version == version])

    return result

def get_compatible_sources():

    # There needs to be an svn library to determine Python version 
    # compatibility because we use the name of the Python .dll or .so file to 
    # determine Python version

    compatible_sources = {}

    matched_version_dict = get_matched_versions()

    for version in [_version for _version in matched_version_dict if 
                    matched_version_dict[_version][0] and 
                    matched_version_dict[_version][1]]:

        for _platform in matched_version_dict[version][1][0].platforms():

            if _platform.os_name is not None and\
               _platform.os_name.casefold() != platform.system().casefold():

                continue

            if _platform.bitness is not None and _platform.bitness != BITNESS:

                continue

            platform_python_versions = _platform.python_versions()

            if sys.version_info[:2] in platform_python_versions:

                compatible_sources[version] = matched_version_dict[version]

    return compatible_sources

def checkout_version(path:pathlib.Path, version: str, makedirs: bool = False):
    """Get all the Blender sources for a specific version
    
    Arguments:
        path {str} -- the path to check out the version
        version {int} -- the version to check out the sources of
    
    Keyword Arguments:
        makedirs {bool} -- automatically make directories (default: {False})
    """

    version = pkg_resources.parse_version(version)

    version_dict = get_matched_versions()

    if version in version_dict:

        if makedirs:

            os.makedirs(path, exist_ok=True)

        else:

            os.mkdir(path)

        # Checkout `git` sources
        version_dict[version][0][0].checkout(path)

        if len(version_dict[version]) == 2:
            # Checkout `svn` sources
            version_dict[version][1][0].checkout(path)

    else:

        raise VersionNotFoundError(f"Version {str(version)} of Blender does "
                                   f"not exist.")

def checkout_all(path: pathlib.Path, makedirs: bool = False):
    """Get all the Blender sources for all versions
    
    Arguments:
        path {str} -- The folder to check it out to
        makedirs {bool} -- Automatically create the checkout directory
    """

    for version, vcs in get_matched_versions().items():

        version_dir = os.path.join(path, str(version))

        if makedirs:

            os.makedirs(version_dir, exist_ok=True)

        else:

            os.mkdir(version_dir)

        # Checkout `git` sources
        vcs[0][0].checkout(version_dir)

        if len(vcs) == 2:
            # Checkout `svn` sources
            vcs[1][0].checkout(path)



#! /usr/bin/python
# -*- coding: future_fstrings -*-
"""
Setup for build scripting
"""

from setuptools import find_packages, setup

setup(name='bpy-build',
      version='1.1.5',
      packages=find_packages(),
      description='Find Blender sources in version control, create build scripts',
      long_description=open("./README.md", 'r').read(),
      long_description_content_type="text/markdown",
      keywords="Blender, 3D, Animation, Renderer, Rendering",
      classifiers=["Development Status :: 3 - Alpha",
                   "Environment :: Win32 (MS Windows)",
                   "Intended Audience :: Developers",
                   "License :: OSI Approved :: "
                   "GNU General Public License v3 (GPLv3)",
                   "Natural Language :: English",
                   "Operating System :: Microsoft :: Windows :: Windows 10",
                   "Programming Language :: C",
                   "Programming Language :: C++",
                   "Programming Language :: Python",
                   "Programming Language :: Python :: 3.4",
                   "Programming Language :: Python :: 3.5",
                   "Programming Language :: Python :: 3.6",
                   "Programming Language :: Python :: 3.7",
                   "Programming Language :: Python :: Implementation :: "
                   "CPython",
                   "Topic :: Artistic Software",
                   "Topic :: Education",
                   "Topic :: Multimedia",
                   "Topic :: Multimedia :: Graphics",
                   "Topic :: Multimedia :: Graphics :: 3D Modeling",
                   "Topic :: Multimedia :: Graphics :: 3D Rendering",
                   "Topic :: Games/Entertainment"],
      author='Tyler Gubala',
      author_email='gubalatyler@gmail.com',
      license='GPL-3.0',
      python_requires=">=3.4.0",
      install_requires=["cmake", "cmake-generators", "GitPython", 'svn'],
      setup_requires=["future-fstrings"],
      url="https://github.com/TylerGubala/bpy-build"
     )
